import logging

from jaeger_client import Config, ConstSampler
from jaeger_client.reporter import NullReporter

from core.jaeger_custom_client.kafka_codec import KafkaCodec, KAFKA_MAP


class ExtendedConfig(Config):
    """
    Конфиг расширяет базовый класс, вводя поддержку KafkaCodec и возможность отключать трасировку
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.logger = logging.getLogger(__name__)

    def initialize_tracer(self, io_loop=None):
        """
        Initialize Jaeger Tracer based on the passed `jaeger_client.Config`.
        Save it to `opentracing.tracer` global variable.
        Only the first call to this method has any effect.
        """

        with Config._initialized_lock:
            if Config._initialized:
                self.logger.warning('Jaeger tracer already initialized, skipping')
                return
            Config._initialized = True

        if self.enabled:
            tracer = self.new_tracer(io_loop)
        else:
            reporter = NullReporter()
            sampler = ConstSampler(False)
            tracer = self.create_tracer(reporter, sampler, None)

        self._initialize_global_tracer(tracer=tracer)
        return tracer

    @property
    def propagation(self):
        codec = KafkaCodec(url_encoding=False,
                           trace_id_header=self.trace_id_header,
                           baggage_header_prefix=self.baggage_header_prefix,
                           debug_id_header=self.debug_id_header)
        return {KAFKA_MAP: codec}



from lazy import lazy


class Device:

    def __init__(self, value):
        self._value = value

    @lazy
    def value(self):
        return self._value

    @lazy
    def platform_type(self):
        return self._value.get("platformType") or ""

    @lazy
    def platform_version(self):
        return self._value.get("platformVersion") or ""

    @lazy
    def surface(self):
        return self._value.get("surface") or ""

    @lazy
    def surface_version(self):
        return self._value.get("surfaceVersion") or ""

    @lazy
    def features(self):
        return self._value.get("features") or {}

    @lazy
    def capabilities(self):
        return self._value.get("capabilities") or {}

    @lazy
    def additional_info(self):
        return self._value.get("additionalInfo") or {}

    @lazy
    def tenant(self):
        return self._value.get("tenant") or ""



from core.names.field import SERVER_ACTION
from core.logging.logger_utils import log
from core.text_preprocessing.preprocessing_result import TextPreprocessingResult

import scenarios.logging.logger_constants as log_const

from smart_kit.handlers.handler_base import HandlerBase
from smart_kit.utils.monitoring import smart_kit_metrics


class HandlerServerAction(HandlerBase):
    handler_name = "HandlerServerAction"

    def __init__(self, app_name, action_name=None, ):
        super(HandlerServerAction, self).__init__(app_name)
        self._action_name = action_name

    def get_action_name(self, payload, user):
        return payload[SERVER_ACTION]["action_id"]

    def get_action_params(self, payload):
        return payload[SERVER_ACTION].get("parameters", {})

    def run(self, payload, user):
        params = {log_const.KEY_NAME: "handling_server_action",
                  "server_action_params": str(self.get_action_params(payload)),
                  "server_action_id": self.get_action_name(payload, user)}
        log("HandlerServerAction %(server_action_id)s started", user, params)

        app_info = user.message.app_info
        smart_kit_metrics.counter_incoming(self.app_name, user.message.message_name, self.__class__.__name__,
                                           user, app_info=app_info)

        action_id = self.get_action_name(payload, user)
        action = user.descriptions["external_actions"][action_id]
        action_params = self.get_action_params(payload)
        return action.run(user, TextPreprocessingResult({}), action_params)



import os
import unittest
from time import time
from unittest.mock import Mock, patch

import smart_kit
from core.basic_models.classifiers.basic_classifiers import ExternalClassifier
from core.basic_models.operators.operators import Operator
from core.basic_models.requirement.basic_requirements import Requirement, CompositeRequirement, AndRequirement, \
    OrRequirement, NotRequirement, RandomRequirement, TopicRequirement, TemplateRequirement, RollingRequirement, \
    TimeRequirement, DateTimeRequirement, IntersectionRequirement, ClassifierRequirement
from core.basic_models.requirement.counter_requirements import CounterValueRequirement, CounterUpdateTimeRequirement
from core.basic_models.requirement.device_requirements import ChannelRequirement
from core.model.registered import registered_factories
from smart_kit.text_preprocessing.local_text_normalizer import LocalTextNormalizer


def patch_get_app_config(mock_get_app_config):
    result = Mock()
    sk_path = os.path.dirname(smart_kit.__file__)
    result.STATIC_PATH = os.path.join(sk_path, 'template/static')
    mock_get_app_config.return_value = result
    result.NORMALIZER = LocalTextNormalizer()
    mock_get_app_config.return_value = result


class MockRequirement:
    def __init__(self, items=None):
        items = items or {}
        self.cond = items.get("cond") or False

    def check(self, text_preprocessing_result, user, params):
        return self.cond


class MockTextNormalizationResult:
    def __init__(self, normalized=None, number_of_numbers=None, currencies_number=None, tokens=None):
        number_of_numbers = number_of_numbers or 0
        currencies_number = currencies_number or 0
        if normalized:
            self.words_tokenized = list(token.get("text") for token in normalized)
            self.words_tokenized_set = set(self.words_tokenized)
        else:
            self.words_tokenized = list()
            self.words_tokenized_set = set()
        self.number_of_numbers = number_of_numbers
        self.currencies_number = currencies_number
        self.tokenized_elements_list = tokens


class MockAmountOperator:
    def __init__(self, items):
        self.amount = items["amount"]

    def compare(self, value):
        return value > self.amount


class MockOperator:
    def __init__(self, amount):
        self.amount = amount

    def compare(self, value):
        return value > self.amount


class EQMockOperator:
    def __init__(self, amount):
        self.amount = amount

    def compare(self, value):
        return value == self.amount


class RequirementTest(unittest.TestCase):
    def test_base(self):
        requirement = Requirement(None)
        assert requirement.check(None, None)

    def test_composite(self):
        registered_factories[Requirement] = MockRequirement
        requirement = CompositeRequirement({"requirements": [
            {"cond": True},
            {"cond": True}
        ]})
        self.assertEqual(len(requirement.requirements), 2)
        self.assertTrue(requirement.check(None, None))

    def test_and_success(self):
        registered_factories[Requirement] = MockRequirement
        requirement = AndRequirement({"requirements": [
            {"cond": True},
            {"cond": True}
        ]})
        self.assertTrue(requirement.check(None, None))

    def test_and_fail(self):
        registered_factories[Requirement] = MockRequirement
        requirement = AndRequirement({"requirements": [
            {"cond": True},
            {"cond": False}
        ]})
        self.assertFalse(requirement.check(None, None))

    def test_or_success(self):
        registered_factories[Requirement] = MockRequirement
        requirement = OrRequirement({"requirements": [
            {"cond": True},
            {"cond": False}
        ]})
        self.assertTrue(requirement.check(None, None))

    def test_or_fail(self):
        registered_factories[Requirement] = MockRequirement
        requirement = OrRequirement({"requirements": [
            {"cond": False},
            {"cond": False}
        ]})
        self.assertFalse(requirement.check(None, None))

    def test_not_success(self):
        registered_factories[Requirement] = MockRequirement
        requirement = NotRequirement({"requirement": {"cond": False}})
        self.assertTrue(requirement.check(None, None))

    def test_not_fail(self):
        registered_factories[Requirement] = MockRequirement
        requirement = NotRequirement({"requirement": {"cond": True}})
        self.assertFalse(requirement.check(None, None))

    def test_channel_success(self):
        user = Mock()
        message = Mock(channel="ch1")
        user.message = message
        requirement = ChannelRequirement({"channels": ["ch1"]})
        text_normalization_result = None
        self.assertTrue(requirement.check(text_normalization_result, user))

    def test_channel_fail(self):
        user = Mock()
        message = Mock(channel="ch2")
        user.message = message
        requirement = ChannelRequirement({"channels": ["ch1"]})
        text_normalization_result = None
        self.assertFalse(requirement.check(text_normalization_result, user))

    def test_random_requirement_true(self):
        requirement = RandomRequirement({"percent": 100})
        self.assertTrue(requirement.check(None, None))

    def test_random_requirement_false(self):
        requirement = RandomRequirement({"percent": 0})
        self.assertFalse(requirement.check(None, None))

    def test_topic_requirement(self):
        requirement = TopicRequirement({"topics": ["test"]})
        user = Mock()
        message = Mock()
        message.topic_key = "test"
        user.message = message
        self.assertTrue(requirement.check(None, user))

    def test_counter_value_requirement(self):
        registered_factories[Operator] = MockAmountOperator
        user = Mock()
        counter = Mock()
        counter.__gt__ = Mock(return_value=True)
        user.counters = {"test": counter}
        requirement = CounterValueRequirement({"operator": {"type": "equal", "amount": 2}, "key": "test"})
        self.assertTrue(requirement.check(None, user))

    def test_counter_time_requirement(self):
        registered_factories[Operator] = MockAmountOperator
        user = Mock()
        counter = Mock()
        counter.update_time = int(time()) - 10
        user.counters = {"test": counter}
        requirement = CounterUpdateTimeRequirement({"operator": {"type": "more_or_equal", "amount": 5}, "key": "test"})
        self.assertTrue(requirement.check(None, user))

    def test_template_req_true(self):
        items = {
            "template": "{{ payload.message.strip() in payload.murexIds }}"
        }
        requirement = TemplateRequirement(items)
        params = {"payload": {
            "groupCode": "BROKER",
            "murexIds": ["AAA", "BBB"],
            "message": " BBB    "
        }}
        user = Mock()
        user.parametrizer = Mock()
        user.parametrizer.collect = Mock(return_value=params)
        self.assertTrue(requirement.check(None, user))

    def test_template_req_false(self):
        items = {
            "template": "{{ payload.groupCode == 'BROKER' }}"
        }
        requirement = TemplateRequirement(items)
        params = {"payload": {"groupCode": "BROKER1"}}
        user = Mock()
        user.parametrizer = Mock()
        user.parametrizer.collect = Mock(return_value=params)
        self.assertFalse(requirement.check(None, user))

    def test_template_req_raise(self):
        items = {
            "template": "{{ payload.groupCode }}"
        }
        requirement = TemplateRequirement(items)
        params = {"payload": {"groupCode": "BROKER1"}}
        user = Mock()
        user.parametrizer = Mock()
        user.parametrizer.collect = Mock(return_value=params)
        self.assertRaises(TypeError, requirement.check, None, user)

    def test_rolling_requirement_true(self):
        user = Mock()
        user.id = "353454"
        requirement = RollingRequirement({"percent": 100})
        text_normalization_result = None
        self.assertTrue(requirement.check(text_normalization_result, user))

    def test_rolling_requirement_false(self):
        user = Mock()
        user.id = "353454"
        requirement = RollingRequirement({"percent": 0})
        text_normalization_result = None
        self.assertFalse(requirement.check(text_normalization_result, user))

    def test_time_requirement_true(self):
        user = Mock()
        user.id = "353454"
        user.message.payload = {
            "meta": {
                "time": {
                    "timestamp": 1610990255000,  # ~ 2021-01-18 17:17:35
                    "timezone_offset_sec": 1000000000,  # shouldn't affect
                }
            }
        }
        requirement = TimeRequirement(
            {
                "operator": {
                    "type": "more",
                    "amount": "17:00:00",
                }
            }
        )
        text_normalization_result = None
        self.assertTrue(requirement.check(text_normalization_result, user))

    def test_time_requirement_false(self):
        user = Mock()
        user.id = "353454"
        user.message.payload = {
            "meta": {
                "time": {
                    "timestamp": 1610979455663,  # ~ 2021-01-18 17:17:35
                    "timezone_offset_sec": 1000000000,  # shouldn't affect
                }
            }
        }
        requirement = TimeRequirement(
            {
                "operator": {
                    "type": "more",
                    "amount": "18:00:00",
                }
            }
        )
        text_normalization_result = None
        self.assertFalse(requirement.check(text_normalization_result, user))

    def test_datetime_requirement_true(self):
        user = Mock()
        user.id = "353454"
        user.message.payload = {
            "meta": {
                "time": {
                    "timestamp": 1610979455663,  # ~ 2021-01-18 17:17:35
                    "timezone_offset_sec": 1000000000,  # shouldn't affect
                }
            }
        }
        requirement = DateTimeRequirement(
            {
                "match_cron": "*/17 14-19 * * mon"
            }
        )
        text_normalization_result = None
        self.assertTrue(requirement.check(text_normalization_result, user))

    def test_datetime_requirement_false(self):
        user = Mock()
        user.id = "353454"
        user.message.payload = {
            "meta": {
                "time": {
                    "timestamp": 1610979455663,  # ~ 2021-01-18 17:17:35
                    "timezone_offset_sec": 1000000000,  # shouldn't affect
                }
            }
        }
        requirement = DateTimeRequirement(
            {
                "match_cron": "* * * * 6,7"
            }
        )
        text_normalization_result = None
        self.assertFalse(requirement.check(text_normalization_result, user))

    @patch('smart_kit.configs.get_app_config')
    def test_intersection_requirement_true(self, mock_get_app_config):
        patch_get_app_config(mock_get_app_config)
        user = Mock()
        requirement = IntersectionRequirement(
            {
                "phrases": [
                    'да',
                    'давай',
                    'хочу',
                ]
            }
        )
        text_normalization_result = Mock()
        text_normalization_result.tokenized_elements_list = [
            {'lemma': 'я'},
            {'lemma': 'хотеть'},
        ]
        self.assertTrue(requirement.check(text_normalization_result, user))

    @patch('smart_kit.configs.get_app_config')
    def test_intersection_requirement_false(self, mock_get_app_config):
        patch_get_app_config(mock_get_app_config)
        user = Mock()
        requirement = IntersectionRequirement(
            {
                "phrases": [
                    'да',
                    'давай',
                    'хочу',
                ]
            }
        )
        text_normalization_result = Mock()
        text_normalization_result.tokenized_elements_list = [
            {'lemma': 'ни'},
            {'lemma': 'за'},
            {'lemma': 'что'},
        ]
        self.assertFalse(requirement.check(text_normalization_result, user))

    @patch.object(ExternalClassifier, "find_best_answer", return_value=[{"answer": "нет", "score": 1.0, "other": False}])
    def test_classifier_requirement_true(self, mock_classifier_model):
        """Тест кейз проверяет что условие возвращает True, если результат классификации запроса относится к одной
        из указанных категорий, прошедших порог, но не равной классу other.
        """
        test_items = {"type": "classifier", "classifier": {"type": "external", "classifier": "hello_scenario_classifier"}}
        classifier_requirement = ClassifierRequirement(test_items)
        mock_user = Mock()
        mock_user.descriptions = {"external_classifiers": ["read_book_or_not_classifier", "hello_scenario_classifier"]}
        result = classifier_requirement.check(Mock(), mock_user)
        self.assertTrue(result)

    @patch.object(ExternalClassifier, "find_best_answer", return_value=[])
    def test_classifier_requirement_false(self, mock_classifier_model):
        """Тест кейз проверяет что условие возвращает False, если модель классификации не вернула ответ."""
        test_items = {"type": "classifier", "classifier": {"type": "external", "classifier": "hello_scenario_classifier"}}
        classifier_requirement = ClassifierRequirement(test_items)
        mock_user = Mock()
        mock_user.descriptions = {"external_classifiers": ["read_book_or_not_classifier", "hello_scenario_classifier"]}
        result = classifier_requirement.check(Mock(), mock_user)
        self.assertFalse(result)

    @patch.object(ExternalClassifier, "find_best_answer", return_value=[{"answer": "other", "score": 1.0, "other": True}])
    def test_classifier_requirement_false_if_class_other(self, mock_classifier_model):
        """Тест кейз проверяет что условие возвращает False, если наиболее вероятный вариант есть класс other."""
        test_items = {"type": "classifier", "classifier": {"type": "external", "classifier": "hello_scenario_classifier"}}
        classifier_requirement = ClassifierRequirement(test_items)
        mock_user = Mock()
        mock_user.descriptions = {"external_classifiers": ["read_book_or_not_classifier", "hello_scenario_classifier"]}
        result = classifier_requirement.check(Mock(), mock_user)
        self.assertFalse(result)


if __name__ == '__main__':
    unittest.main()



import inspect
import logging
from typing import Dict, List, Union, Optional

import timeout_decorator

import core.basic_models.classifiers.classifiers_constants as cls_const
import core.logging.logger_constants as log_const
import scenarios.logging.logger_constants as scenarios_log_const
from core.basic_models.classifiers.basic_classifiers import Classifier
from core.logging.masker import LogMasker
from core.utils.stats_timer import StatsTimer

MESSAGE_ID_STR = "message_id"
UID_STR = "uid"
LOGGING_UUID = "logging_uuid"
CLASS_NAME = "class_name"


def _make_message(user=None, params=None, cls_name=''):
    message_id = None
    uuid = None
    logging_uuid = None
    params = params or {}
    if user:
        message = user.message
        message_id = message.incremental_id
        uuid = user.id
        logging_uuid = message.logging_uuid
        atr_names = ["channel", "type", "device_channel", "device_channel_version", "device_platform", "group",
                     "device_platform_version", "device_platform_client_type", "csa_profile_id", "test_deploy"]
        for name in atr_names:
            param = getattr(message, name, None)
            if param:
                params[name] = param
    params = LogMasker.mask_structure(params, LogMasker.percent_fix)
    params[UID_STR] = uuid
    params[MESSAGE_ID_STR] = message_id
    params[LOGGING_UUID] = logging_uuid
    params[CLASS_NAME] = cls_name
    return params


app_logger = logging.getLogger(log_const.APP_LOGGER_NAME)


def log(message, user=None, params=None, level="INFO", exc_info=None):
    try:
        level_name = logging.getLevelName(level)
        current_frame = inspect.currentframe()
        previous_frame = current_frame.f_back.f_locals
        instance = previous_frame.get('self', None)
        if instance is not None:
            params = _make_message(user, params, instance.__class__.__name__)
        else:
            params = _make_message(user, params)

        masked_message = LogMasker.mask_structure(message, LogMasker.regular_exp)
        masked_params = LogMasker.mask_structure(params, LogMasker.regular_exp)

        app_logger.log(level_name, masked_message, masked_params, exc_info=exc_info)
    except timeout_decorator.TimeoutError:
        raise
    except:
        app_logger.log(logging.getLevelName("ERROR"), "Failed to write a log. Exception occurred",
                       params, exc_info=True)


def log_classifier_result(classification_res: List[Dict[str, Union[str, float, bool]]], user,
                          classifier: Classifier, timer: Optional[StatsTimer] = None) -> None:
    classifier_name = classifier.settings.get("classifier", "intent_recognizer")
    score_key = cls_const.INTENT_RECOGNIZER_ANSWER_DISTANCE_KEY if classifier_name == "intent_recognizer" \
        else cls_const.SCORE_KEY
    params = {
        log_const.KEY_NAME: scenarios_log_const.CLASSIFIER_VALUE,
        "classifier_name": classifier_name,
        "result": [el[cls_const.ANSWER_KEY] for el in classification_res],
        "score": [el[score_key] for el in classification_res]
    }
    if timer:
        params["time"] = timer.msecs

    log(scenarios_log_const.CLASSIFIER_MESSAGE, user, params)


# backward naming_compatibility
behaviour_log = log



import requests

from typing import Optional, Dict, Union, List

from core.basic_models.actions.command import Command
from core.basic_models.actions.string_actions import NodeAction
from core.model.base_user import BaseUser
from core.text_preprocessing.base import BaseTextPreprocessingResult


class HTTPRequestAction(NodeAction):
    def __init__(self, items, id=None):
        super().__init__(items, id)
        self.params = items["params"]
        self.url = self.params["url"]
        self.method = self.params["method"]
        self.store = items["store"]
        self.behavior = items["behavior"]

    def run(self, user: BaseUser, text_preprocessing_result: BaseTextPreprocessingResult,
            params: Optional[Dict[str, Union[str, float, int]]] = None) -> Optional[List[Command]]:
        behavior_description = user.descriptions["behaviors"][self.behavior]

        request_params = self.params
        request_params["timeout"] = behavior_description.timeout(user)

        params = params or {}
        collected = user.parametrizer.collect(text_preprocessing_result)
        params.update(collected)

        request_parameters = self._get_rendered_tree_recursive(self._get_template_tree(request_params), params)

        try:
            with requests.request(**request_parameters) as response:
                response.raise_for_status()
                user.variables.set(self.store, response.json())
                return behavior_description.success_action.run(user, text_preprocessing_result, None)
        except requests.exceptions.Timeout:
            return behavior_description.timeout_action.run(user, text_preprocessing_result, None)
        except (requests.exceptions.ConnectionError, requests.exceptions.HTTPError):
            return behavior_description.fail_action.run(user, text_preprocessing_result, None)



import collections
import json
import re
from itertools import islice
from typing import Dict, List, Union, Optional, Any, Callable, Pattern, Set

from jinja2 import exceptions as jexcept
from lazy import lazy

import core.basic_models.classifiers.classifiers_constants as cls_const
import core.logging.logger_constants as core_log_const
import scenarios.logging.logger_constants as log_const
from core.basic_models.classifiers.basic_classifiers import Classifier, ExternalClassifier
from core.logging.logger_utils import log, log_classifier_result
from core.model.factory import build_factory
from core.model.factory import factory
from core.model.factory import list_factory
from core.model.registered import Registered
from core.text_preprocessing.base import BaseTextPreprocessingResult
from core.text_preprocessing.preprocessing_result import TextPreprocessingResult
from core.unified_template.unified_template import UnifiedTemplate
from core.utils.exception_handlers import exc_handler
from core.utils.pickle_copy import pickle_deepcopy
from core.utils.stats_timer import StatsTimer
from scenarios.user.user_model import User

field_filler_description = Registered()

field_filler_factory = build_factory(field_filler_description)


class FieldFillerDescription:

    def __init__(self, items: Optional[Dict[str, Any]], id: Optional[str] = None) -> None:
        items = items or {}
        self.id = id
        self.version = items.get("version", -1)

    def _log_params(self):
        return {
            log_const.KEY_NAME: log_const.FILLER_EXTRACT_VALUE,
            "filler": self.__class__.__name__
        }

    def extract(self, text_preprocessing_result: BaseTextPreprocessingResult, user: User,
                params: Dict[str, Any] = None) -> None:
        return None

    def on_extract_error(self, text_preprocessing_result, user, params=None):
        log("exc_handler: Filler failed to extract. Return None. MESSAGE: {}.".format(user.message.masked_value), user,
            {log_const.KEY_NAME: core_log_const.HANDLED_EXCEPTION_VALUE}, level="ERROR", exc_info=True)
        return None

    def run(self, user: User, text_preprocessing_result: BaseTextPreprocessingResult,
            params: Optional[Dict[str, Any]] = None) -> None:
        return self.extract(text_preprocessing_result, user, params)

    def _postprocessing(self, user: User, item: str) -> None:
        last_scenario_name = user.last_scenarios.last_scenario_name
        user.scenario_models[last_scenario_name].postprocessing_data.append(item)


class ExternalFieldFillerDescription(FieldFillerDescription):
    filler: Optional[str]

    def __init__(self, items: Optional[Dict[str, Any]], id: Optional[str] = None) -> None:
        super(ExternalFieldFillerDescription, self).__init__(items, id)
        self.filler = items.get("filler")

    # @exc_handler(on_error_obj_method_name="on_extract_error")
    def extract(self, text_preprocessing_result: BaseTextPreprocessingResult,
                user: User, params: Dict[str, Any] = None) -> Optional[Union[int, float, str, bool, List, Dict]]:
        filler = user.descriptions["external_field_fillers"][self.filler]
        return filler.run(user, text_preprocessing_result, params)


class CompositeFiller(FieldFillerDescription):
    fillers: Optional[List[FieldFillerDescription]]

    def __init__(self, items: Optional[Dict[str, Any]], id: Optional[str] = None) -> None:
        super(CompositeFiller, self).__init__(items, id)
        self._fillers: Optional[List[Dict[str, Any]]] = items.get("fillers") or []

    @lazy
    @list_factory(FieldFillerDescription)
    def fillers(self):
        return self._fillers

    @exc_handler(on_error_obj_method_name="on_extract_error")
    def extract(self, text_preprocessing_result: BaseTextPreprocessingResult,
                user: User, params: Dict[str, Any] = None) -> Optional[Union[int, float, str, bool, List, Dict]]:
        extracted = None
        for filler in self.fillers:
            extracted = filler.extract(text_preprocessing_result, user, params)
            if extracted is not None:
                break
        return extracted


class AvailableInfoFiller(FieldFillerDescription):
    loader: Optional[str]
    value: Union[str, Dict]
    loaders = collections.defaultdict(str, {"json": json.loads, "float": float, "int": int})

    def __init__(self, items: Optional[Dict[str, Any]], id: Optional[str] = None) -> None:
        super(AvailableInfoFiller, self).__init__(items, id)
        value = items['value']
        self.loader = items.get('loader')
        self.template: UnifiedTemplate = UnifiedTemplate(value)

    @exc_handler(on_error_obj_method_name="on_extract_error")
    def extract(self, text_preprocessing_result: BaseTextPreprocessingResult,
                user: User, params: Dict[str, Any] = None) -> Optional[Union[int, float, str, bool, List, Dict]]:
        params = params or {}
        collected = user.parametrizer.collect(text_preprocessing_result)
        params.update(collected)
        try:
            # if path is wrong, it may fail with UndefinedError
            # notion: {key: None} will return "None";
            # not existing key or value "" will return ""; otherwise question in scenario will go in cycles
            value = self.template.silent_render(params)
        except jexcept.UndefinedError:
            value = None

        if self.loader:
            if value:
                loader = self.loaders[self.loader]
                value = loader(value)
            else:
                value = None

        return value


class FirstNumberFiller(FieldFillerDescription):
    @exc_handler(on_error_obj_method_name="on_extract_error")
    def extract(self, text_preprocessing_result: BaseTextPreprocessingResult, user: User,
                params: Dict[str, Any] = None) -> Optional[int]:
        numbers = text_preprocessing_result.num_token_values
        if numbers:
            log_params = self._log_params()
            log_params["numbers"] = str(numbers)
            message = "Filler: %(filler)s, Numbers: %(numbers)s"
            log(message, user, log_params)
        return numbers[0] if numbers else None


class FirstCurrencyFiller(FieldFillerDescription):

    @exc_handler(on_error_obj_method_name="on_extract_error")
    def extract(self, text_preprocessing_result: BaseTextPreprocessingResult, user: User,
                params: Dict[str, Any] = None) -> Optional[str]:
        currencies = text_preprocessing_result.ccy_token_values
        if currencies:
            log_params = self._log_params()
            log_params["currencies"] = str(currencies)
            message = "Filler: %(filler)s, Currencies: %(currencies)s"
            log(message, user, log_params)
        return currencies[0] if currencies else None


class FirstOrgFiller(FieldFillerDescription):

    @exc_handler(on_error_obj_method_name="on_extract_error")
    def extract(self, text_preprocessing_result: BaseTextPreprocessingResult, user: User,
                params: Dict[str, Any] = None) -> Optional[str]:
        orgs = text_preprocessing_result.org_token_values
        if orgs:
            log_params = self._log_params()
            log_params["orgs"] = str(orgs)
            message = "Filler: %(filler)s, Organisations: %(orgs)s"
            log(message, user, log_params)
        return orgs[0] if orgs else None


class FirstGeoFiller(FieldFillerDescription):

    @exc_handler(on_error_obj_method_name="on_extract_error")
    def extract(self, text_preprocessing_result: BaseTextPreprocessingResult, user: User,
                params: Dict[str, Any] = None) -> Optional[str]:
        geos = text_preprocessing_result.geo_token_values
        if geos:
            log_params = self._log_params()
            log_params["geos"] = str(geos)
            message = "Filler: %(filler)s, Toponyms: %(geos)s"
            log(message, user, log_params)
        return geos[0] if geos else None


class RegexpFieldFiller(FieldFillerDescription):
    regexp: str
    delimiter: Optional[str]

    def __init__(self, items: Optional[Dict[str, Any]], id: Optional[str] = None) -> None:
        super(RegexpFieldFiller, self).__init__(items, id)
        self.regexp = items["exp"]
        self.delimiter = items.get("delimiter", ",")

    @exc_handler(on_error_obj_method_name="on_extract_error")
    def extract(self, text_preprocessing_result: BaseTextPreprocessingResult, user: User,
                params: Dict[str, Any] = None) -> Optional[str]:
        original_text = text_preprocessing_result.original_text
        match = re.findall(self.regexp, original_text)
        if match:
            log_params = self._log_params()
            log_params["match"] = str(match)
            message = "Filler: %(filler)s, Data: %(match)s"
            log(message, user, log_params)
            return self.delimiter.join(match)


class RegexpAndStringOperationsFieldFiller(RegexpFieldFiller):
    regexp: str
    delimiter: Optional[str]
    operations: List[Dict]

    def __init__(self, items: Optional[Dict[str, Any]], id: Optional[str] = None) -> None:
        super(RegexpAndStringOperationsFieldFiller, self).__init__(items, id)
        self.operations = items["operations"]
        self.functions_mapping: Dict[str, Callable] = {
            "strip": str.strip,
            "rstrip": str.rstrip,
            "lstrip": str.lstrip,
            "upper": str.upper,
            "lower": str.lower
        }

    def _operation(self, original_text, typeOp, amount):
        func = self.functions_mapping[typeOp]
        return func(original_text, amount) if amount else func(original_text)

    @exc_handler(on_error_obj_method_name="on_extract_error")
    def extract(self, text_preprocessing_result: BaseTextPreprocessingResult, user: User,
                params: Dict[str, Any] = None) -> Optional[str]:
        original_text = text_preprocessing_result.original_text
        if self.operations:
            for op in self.operations:
                original_text = self._operation(original_text, op["type"], op.get("amount"))
        text_preprocessing_result_copy = pickle_deepcopy(text_preprocessing_result)
        text_preprocessing_result_copy.original_text = original_text
        return super(RegexpAndStringOperationsFieldFiller, self).extract(text_preprocessing_result_copy, user, params)


class AllRegexpsFieldFiller(FieldFillerDescription):
    exps: Optional[List[str]]
    delimiter: Optional[str]
    original_text_lower: Optional[bool]

    def __init__(self, items: Optional[Dict[str, Any]], id: Optional[str] = None) -> None:
        super(AllRegexpsFieldFiller, self).__init__(items, id)
        self.exps = items.get("exps") or []
        self.regexps: List[Pattern] = [re.compile(r) for r in self.exps]
        self.delimiter = items.get("delimiter") or ","
        self.original_text_lower = items.get("original_text_lower") or False

    @exc_handler(on_error_obj_method_name="on_extract_error")
    def extract(self, text_preprocessing_result: BaseTextPreprocessingResult, user: User,
                params: Dict[str, Any] = None) -> Optional[str]:
        original_text = text_preprocessing_result.original_text
        if self.original_text_lower:
            original_text = original_text.lower()
        matches = []
        for r in self.regexps:
            matches.extend(r.findall(original_text))
        if matches:
            result = self.delimiter.join(matches)
            log_params = self._log_params()
            log_params["regexp_result"] = str(result)
            message = "Filler: %(filler)s, Data: %(regexp_result)s"
            log(message, user, log_params)
            return result


class FirstPersonFiller(FieldFillerDescription):

    @exc_handler(on_error_obj_method_name="on_extract_error")
    def extract(self, text_preprocessing_result: BaseTextPreprocessingResult, user: User,
                params: Dict[str, Any] = None) -> Optional[str]:
        persons = text_preprocessing_result.person_token_values
        if persons:
            log_params = self._log_params()
            log_params["persons"] = str(persons)
            message = "Filler: %(filler)s, Persons: %(persons)s"
            log(message, user, log_params)
        return persons[0] if persons else None


class PreviousMessagesFiller(FieldFillerDescription):
    filler: Optional[FieldFillerDescription]
    count: Optional[int]

    def __init__(self, items: Optional[Dict[str, Any]], id: Optional[str] = None) -> None:
        super(PreviousMessagesFiller, self).__init__(items, id)
        self._filler: Optional[Dict[str, Any]] = items.get("filler")
        self.count = items.get("count")

    @lazy
    @factory(FieldFillerDescription)
    def filler(self):
        return self._filler

    @exc_handler(on_error_obj_method_name="on_extract_error")
    def extract(self, text_preprocessing_result: BaseTextPreprocessingResult, user: User,
                params: Dict[str, Any] = None) -> Optional[str]:
        result = self.filler.extract(text_preprocessing_result, user, params)
        if result is None:
            result = self._try_extract_last_messages(user, params)
        return result

    def _try_extract_last_messages(self, user, params):
        processed_items = user.preprocessing_messages_for_scenarios.processed_items
        count = self.count - 1 if self.count else len(processed_items)
        for preprocessing_result_raw in islice(processed_items, 0, count):
            preprocessing_result = TextPreprocessingResult(preprocessing_result_raw)
            result = self.filler.extract(preprocessing_result, user, params)
            if result is not None:
                return result


class UserIdFiller(FieldFillerDescription):

    @exc_handler(on_error_obj_method_name="on_extract_error")
    def extract(self, text_preprocessing_result: BaseTextPreprocessingResult, user: User,
                params: Dict[str, Any] = None) -> Optional[str]:
        result = user.message.uuid.get('userId')
        return result


class IntersectionFieldFiller(FieldFillerDescription):
    cases: Optional[Dict[str, List[str]]]

    def __init__(self, items: Optional[Dict[str, Any]], id: Optional[str] = None) -> None:
        super(IntersectionFieldFiller, self).__init__(items, id)
        self.cases = items.get("cases") or {}
        self.default = items.get("default")
        from smart_kit.configs import get_app_config
        app_config = get_app_config()

        self.normalized_cases = []
        for key, val in self.cases.items():
            tokens_list = []
            for message in app_config.NORMALIZER.normalize_sequence(val):
                case = set()
                for norm in message["tokenized_elements_list"]:
                    if norm.get("token_type") != "SENTENCE_ENDPOINT_TOKEN":
                        case.add(norm.get("lemma"))
                tokens_list.append(case)
            self.normalized_cases.append((key, tokens_list))

    @exc_handler(on_error_obj_method_name="on_extract_error")
    def extract(self, text_preprocessing_result: TextPreprocessingResult, user: User,
                params: Dict[str, Any] = None) -> Optional[str]:
        tpr_tokenized_set = {norm.get("lemma") for norm in text_preprocessing_result.tokenized_elements_list if
                             norm.get("token_type") != "SENTENCE_ENDPOINT_TOKEN"}
        for key, tokens_list in self.normalized_cases:
            for tokens in tokens_list:
                if tpr_tokenized_set >= tokens:
                    log_params = self._log_params()
                    log_params["words_tokenized_set"] = str(tpr_tokenized_set)
                    log_params["tokens"] = str(tokens)
                    message = "Filler: %(filler)s, words_normalized_set: %(words_tokenized_set)s, tokens: %(tokens)s"
                    log(message, user, log_params)
                    return key
        if self.default:
            return self.default


class IntersectionOriginalTextFiller(FieldFillerDescription):

    def __init__(self, items: Optional[Dict[str, Any]], id: Optional[str] = None) -> None:
        super(IntersectionOriginalTextFiller, self).__init__(items, id)
        self.cases = items.get("cases", {})
        self.original_cases = [(key, [{*phrase.split()} for phrase in val]) for key, val in self.cases.items()]

        self._exceptions = items.get("exceptions", {})
        self.exceptions = {key: [{*phrase.split()} for phrase in val] for key, val in self._exceptions.items()}

    def _check_exceptions(self, key, tpr_original_set):
        if key in self.exceptions.keys():
            for exc_tokens in self.exceptions[key]:
                if tpr_original_set >= exc_tokens:
                    return True
        return False

    @exc_handler(on_error_obj_method_name="on_extract_error")
    def extract(self, text_preprocessing_result: TextPreprocessingResult, user: User,
                params: Dict[str, Any] = None) -> Optional[str]:
        tpr_original_set = {*text_preprocessing_result.original_text.split()}
        for key, tokens_list in self.original_cases:
            for tokens in tokens_list:
                if tpr_original_set >= tokens and not self._check_exceptions(key, tpr_original_set):
                    log_params = self._log_params()
                    log_params["tpr_original_set"] = str(tpr_original_set)
                    log_params["tokens"] = str(tokens)
                    message = "Filler: %(filler)s, tpr_original_set: %(tpr_original_set)s, tokens: %(tokens)s"
                    log(message, user, log_params)
                    return key


class ApproveFiller(FieldFillerDescription):
    yes_words: Optional[List]
    no_words: Optional[List]

    def __init__(self, items: Optional[Dict[str, Any]], id: Optional[str] = None) -> None:
        super(ApproveFiller, self).__init__(items, id)

        from smart_kit.configs import get_app_config
        app_config = get_app_config()

        self.yes_words = items.get("yes_words")
        self.no_words = items.get("no_words")
        self.set_yes_words: Set = set(self.yes_words or [])
        self.set_no_words: Set = set(self.no_words or [])
        self.yes_words_normalized: Set = {
            TextPreprocessingResult(result).tokenized_string for result in
            app_config.NORMALIZER.normalize_sequence(self.set_yes_words)
        }
        self.no_words_normalized: Set = {
            TextPreprocessingResult(result).tokenized_string for result in
            app_config.NORMALIZER.normalize_sequence(self.set_no_words)
        }

    @exc_handler(on_error_obj_method_name="on_extract_error")
    def extract(self, text_preprocessing_result: TextPreprocessingResult, user: User,
                params: Dict[str, Any] = None) -> Optional[bool]:
        if text_preprocessing_result.tokenized_string in self.yes_words_normalized:
            params = self._log_params()
            params["tokenized_string"] = text_preprocessing_result.tokenized_string
            params["yes_words"] = self.yes_words_normalized
            message = "Filler: %(filler)s, normalized_text: %(tokenized_string)s, self.yes_words: %(yes_words)s"
            log(message, user, params)
            response = True
        elif text_preprocessing_result.words_tokenized_set.intersection(self.no_words_normalized):
            params = self._log_params()
            params["tokenized_string"] = text_preprocessing_result.words_tokenized_set
            params["no_words"] = self.no_words_normalized
            message = "Filler: %(filler)s, normalized_text: %(tokenized_string)s, self.no_words: %(no_words)s"
            log(message, user, params)
            response = False
        else:
            response = None
        return response


class ApproveRawTextFiller(ApproveFiller):
    @exc_handler(on_error_obj_method_name="on_extract_error")
    def extract(
            self, text_preprocessing_result: TextPreprocessingResult, user: User, params: Dict[str, Any] = None
    ) -> Optional[bool]:
        original_text = ' '.join(text_preprocessing_result.original_text.split()).lower().rstrip('!.)')
        if original_text in self.set_yes_words:
            params = self._log_params()
            params["original_text"] = original_text
            params["yes_words"] = self.set_yes_words
            message = "Filler: %(filler)s, original_text: %(original_text)s, self.yes_words: %(yes_words)s"
            log(message, user, params)
            response = True
        elif text_preprocessing_result.words_tokenized_set.intersection(self.no_words_normalized):
            params = self._log_params()
            params["words_tokenized_set"] = text_preprocessing_result.words_tokenized_set
            params["no_words"] = self.set_no_words
            message = "Filler: %(filler)s, words_normalized_set: %(words_tokenized_set)s, self.no_words: %(no_words)s"
            log(message, user, params)
            response = False
        else:
            response = None
        return response


class ClassifierFiller(FieldFillerDescription):
    """Заполняет поле одним из возможных значений, которое выдает модель классификации.
    Запрос клиента проходит классификацию на основе внешнего классификатора, после чего в качестве ответа
    берётся класс с максимальной вероятностью.
    """

    def __init__(self, items: Optional[Dict[str, Any]], id: Optional[str] = None) -> None:
        super(ClassifierFiller, self).__init__(items, id)
        self._classifier = items["classifier"]
        self._cls_const_answer_key = cls_const.ANSWER_KEY

    @lazy
    def classifier(self) -> Classifier:
        return ExternalClassifier(self._classifier)

    def _get_result(self, answers: List[Dict[str, Union[str, float, bool]]]) -> str:
        return answers[0][self._cls_const_answer_key]

    @exc_handler(on_error_obj_method_name="on_extract_error")
    def extract(self, text_preprocessing_result: BaseTextPreprocessingResult, user: User,
                params: Dict[str, Any] = None) -> Union[str, None, List[Dict[str, Union[str, float, bool]]]]:
        result = None
        classifier = self.classifier
        with StatsTimer() as timer:
            classification_res = classifier.find_best_answer(
                text_preprocessing_result, scenario_classifiers=user.descriptions["external_classifiers"])

        log_classifier_result(classification_res, user, classifier, timer)

        if classification_res:
            params = self._log_params()
            params["answers"] = classification_res
            message = "Filler: %(filler)s, answers: %(answers)s, "
            log(message, user, params)
            result = self._get_result(classification_res)

        return result


class ClassifierFillerMeta(ClassifierFiller):

    def _get_result(self, answers: List[Dict[str, Union[str, float, bool]]]) -> List[Dict[str, Union[str, float, bool]]]:
        return answers



import hashlib
from datetime import datetime, timezone
from random import random
from typing import List, Optional, Dict, Any

from croniter import croniter
from lazy import lazy

import core.logging.logger_constants as log_const
from core.basic_models.classifiers.basic_classifiers import Classifier, ExternalClassifier
from core.basic_models.operators.operators import Operator
from core.logging.logger_utils import log, log_classifier_result
from core.model.base_user import BaseUser
from core.model.factory import build_factory, list_factory, factory
from core.model.registered import Registered
from core.text_preprocessing.base import BaseTextPreprocessingResult
from core.text_preprocessing.preprocessing_result import TextPreprocessingResult
from core.unified_template.unified_template import UnifiedTemplate
from core.utils.stats_timer import StatsTimer
from scenarios.scenario_models.field.field_filler_description import IntersectionFieldFiller
from scenarios.user.user_model import User

requirements = Registered()

requirement_factory = build_factory(requirements)


class Requirement:
    # should_process_message compatible
    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        items = items or {}
        self.items = items
        self.version = items.get("version", -1)
        self.id = id

    def _log_params(self):
        return {
            log_const.KEY_NAME: log_const.REQUIREMENT_CHECK_VALUE,
            "requirement": self.__class__.__name__
        }

    def check(self, text_preprocessing_result: BaseTextPreprocessingResult, user: BaseUser,
              params: Dict[str, Any] = None) -> bool:
        return True

    def on_check_error(self, text_preprocessing_result, user):
        log("exc_handler: Requirement failed to check. Return False. MESSAGE: {}.".format(user.message.masked_value),
            user, {log_const.KEY_NAME: log_const.HANDLED_EXCEPTION_VALUE},
            level="ERROR", exc_info=True)
        return False


class CompositeRequirement(Requirement):
    requirements: List[Requirement]

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super(CompositeRequirement, self).__init__(items, id)
        self._requirements = items["requirements"]

    @lazy
    @list_factory(Requirement)
    def requirements(self):
        return self._requirements


class AndRequirement(CompositeRequirement):

    def check(self, text_preprocessing_result: BaseTextPreprocessingResult, user: BaseUser,
              params: Dict[str, Any] = None) -> bool:
        return all(requirement.check(text_preprocessing_result=text_preprocessing_result, user=user, params=params)
                   for requirement in self.requirements)


class OrRequirement(CompositeRequirement):

    def check(self, text_preprocessing_result: BaseTextPreprocessingResult, user: BaseUser,
              params: Dict[str, Any] = None) -> bool:
        return any(requirement.check(text_preprocessing_result=text_preprocessing_result, user=user, params=params)
                   for requirement in self.requirements)


class NotRequirement(Requirement):
    requirement: Requirement

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super(NotRequirement, self).__init__(items, id)
        self._requirement = items["requirement"]

    @lazy
    @factory(Requirement)
    def requirement(self):
        return self._requirement

    def check(self, text_preprocessing_result: BaseTextPreprocessingResult, user: BaseUser,
              params: Dict[str, Any] = None) -> bool:
        return not self.requirement.check(text_preprocessing_result=text_preprocessing_result, user=user, params=params)


class ComparisonRequirement(Requirement):
    operator: Operator

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super(ComparisonRequirement, self).__init__(items, id)
        self._operator = items["operator"]

    @lazy
    @factory(Operator)
    def operator(self):
        return self._operator


class RandomRequirement(Requirement):
    percent: int

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super(RandomRequirement, self).__init__(items, id)
        self.percent = items["percent"]

    def check(self, text_preprocessing_result: BaseTextPreprocessingResult, user: BaseUser,
              params: Dict[str, Any] = None) -> bool:
        result = random() * 100
        return result < self.percent


class TopicRequirement(Requirement):
    topics: List[str]

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super(TopicRequirement, self).__init__(items, id)
        self.topics = items["topics"]

    def check(self, text_preprocessing_result: BaseTextPreprocessingResult, user: BaseUser,
              params: Dict[str, Any] = None) -> bool:
        return user.message.topic_key in self.topics


class TemplateRequirement(Requirement):
    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super(TemplateRequirement, self).__init__(items, id)
        self._template = UnifiedTemplate(items["template"])

    def check(self, text_preprocessing_result: BaseTextPreprocessingResult, user: BaseUser,
              params: Dict[str, Any] = None) -> bool:
        params = params or {}
        collected = user.parametrizer.collect(text_preprocessing_result)
        params.update(collected)
        render_result = self._template.render(params)
        if render_result == "True":
            return True
        if render_result == "False":
            return False
        raise TypeError(f'Template result should be "True" or "False", got: ',
                        f'{render_result} for template {self.items["template"]}')


class RollingRequirement(Requirement):
    percent: int

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super(RollingRequirement, self).__init__(items, id)
        self.percent = items["percent"]

    def check(self, text_preprocessing_result: BaseTextPreprocessingResult, user: BaseUser,
              params: Dict[str, Any] = None) -> bool:
        id = user.id
        s = id.encode('utf-8')
        hash = int(hashlib.sha256(s).hexdigest(), 16)
        res = hash % 100
        return res < self.percent


class TimeRequirement(ComparisonRequirement):
    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super().__init__(items, id)

    def check(
            self,
            text_preprocessing_result: BaseTextPreprocessingResult,
            user: BaseUser,
            params: Dict[str, Any] = None
    ) -> bool:
        message_time_dict = user.message.payload['meta']['time']
        message_timestamp_sec = message_time_dict['timestamp'] // 1000
        message_time = datetime.fromtimestamp(message_timestamp_sec, tz=timezone.utc).time()
        return self.operator.compare(message_time)

    @lazy
    @factory(Operator)
    def operator(self):
        operator = dict(self._operator)
        amount_time = datetime.strptime(operator["amount"], '%H:%M:%S').time()
        operator["amount"] = amount_time
        return operator


class DateTimeRequirement(Requirement):
    match_cron: str

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super().__init__(items, id)
        self.match_cron = items['match_cron']

    def check(
            self,
            text_preprocessing_result: BaseTextPreprocessingResult,
            user: BaseUser,
            params: Dict[str, Any] = None
    ) -> bool:
        message_time_dict = user.message.payload['meta']['time']
        message_timestamp_sec = message_time_dict['timestamp'] // 1000
        message_datetime = datetime.fromtimestamp(message_timestamp_sec)
        return croniter.match(self.match_cron, message_datetime)


class IntersectionRequirement(Requirement):
    phrases: Optional[List]

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super().__init__(items, id)
        self.filler = IntersectionFieldFiller(
            {
                'cases': {
                    True: items.get('phrases', []),
                },
                'default': False,
            },
            id,
        )

    def check(
            self,
            text_preprocessing_result: TextPreprocessingResult,
            user: User,
            params: Dict[str, Any] = None
    ) -> bool:
        result = bool(
            self.filler.extract(text_preprocessing_result, user, params),
        )
        return result


class ClassifierRequirement(Requirement):
    """Условие, которое зависит от результата классификации.
    Возвращает True, если результат классификации запроса относится к одной из указанных категорий, прошедших порог,
    но не равной классу other.
    """

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super(ClassifierRequirement, self).__init__(items=items, id=id)
        self._classifier = items["classifier"]

    @lazy
    def classifier(self) -> Classifier:
        return ExternalClassifier(self._classifier)

    def check(self, text_preprocessing_result: BaseTextPreprocessingResult, user: BaseUser,
              params: Dict[str, Any] = None) -> bool:
        check_res = True
        classifier = self.classifier
        with StatsTimer() as timer:
            classification_res = classifier.find_best_answer(
                text_preprocessing_result, scenario_classifiers=user.descriptions["external_classifiers"])

        log_classifier_result(classification_res, user, classifier, timer)

        if not classification_res or classification_res[0][classifier.class_other]:
            check_res = False

        return check_res



from typing import Dict, Any

from confluent_kafka.cimpl import Message
from opentracing import Span

from core.jaeger_custom_client.jaeger_config import KAFKA_MAP
from core.message.from_message import SmartAppFromMessage
import core.message.message_constants as const


def get_incoming_spam(tracer, from_message: SmartAppFromMessage, mq_message: Message) -> Span:
    """
    Метод извлекает информацию об trace из входящего сообщения кафки. Созданный Span
    обогащается тегами
    :param tracer: Трейсер
    :param from_message: Распарщенное сообщение
    :param mq_message: Сообщение из кафки
    :return: Проинициализированный и активированный Span
    """
    span_context = tracer.extract(KAFKA_MAP, mq_message)
    span = tracer.start_span(
        operation_name=from_message.type,
        child_of=span_context)
    params = build_log_params(from_message)
    for key, value in params.items():
        span.set_tag(key, value)
    return span


def build_log_params(from_message: SmartAppFromMessage) -> Dict[str, Any]:
    logging_dict = {
        const.MSG_MESSAGEID_KEY: from_message.incremental_id,
        const.MSG_USERCHANNEL_KEY: from_message.channel,
        const.MSG_USERID_KEY: from_message.uid
    }
    chat_id = getattr(from_message, "chat_id", None)
    if chat_id:
        logging_dict[const.MSG_CHATID_KEY] = chat_id
    return logging_dict



# coding: utf-8


class Command:
    def __init__(self, name, params=None, action_id=None, request_type=None, request_data=None):
        self.name = name
        self.payload = params or {}
        self.action_id = action_id
        self.request_type = request_type
        self.request_data = request_data or {}

    @property
    def raw(self):
        message = {"messageName": self.name, "payload": self.payload}
        if self.action_id is not None:
            message["action_id"] = self.action_id
        return message



# coding: utf-

import scenarios.logging.logger_constants as log_const
from core.logging.logger_utils import log
from core.text_preprocessing.preprocessing_result import TextPreprocessingResult

from smart_kit.handlers.handler_base import HandlerBase


class HandlerText(HandlerBase):

    def __init__(self, app_name, dialogue_manager):
        super(HandlerText, self).__init__(app_name)
        log(
            f"{self.__class__.__name__}.__init__ started.", params={log_const.KEY_NAME: log_const.STARTUP_VALUE}
        )
        self.dialogue_manager = dialogue_manager
        log(
            f"{self.__class__.__name__}.__init__ finished.", params={log_const.KEY_NAME: log_const.STARTUP_VALUE}
        )

    def run(self, payload, user):
        super().run(payload, user)
        text_preprocessing_result = TextPreprocessingResult(payload.get("message", {}))

        log("text preprocessing result: '{}'".format(str(text_preprocessing_result.raw)), user,
                      {log_const.KEY_NAME: log_const.NORMALIZED_TEXT_VALUE})

        answer = self._handle_base(text_preprocessing_result, user)
        return answer

    def _handle_base(self, text_preprocessing_result, user):
        answer, is_answer_found = self.dialogue_manager.run(text_preprocessing_result, user)
        return answer or []



import json
import os
from typing import AnyStr, Optional, Tuple, Any, Dict

from lazy import lazy

from core.configs.global_constants import LINK_BEHAVIOR_FLAG, CALLBACK_ID_HEADER
from core.message.from_message import SmartAppFromMessage
from smart_kit.compatibility.commands import combine_commands
from smart_kit.configs import Settings
from smart_kit.message.smartapp_to_message import SmartAppToMessage
from smart_kit.models.smartapp_model import SmartAppModel
from smart_kit.request.kafka_request import SmartKitKafkaRequest
from smart_kit.testing.utils import Environment
from smart_kit.utils.diff import partial_diff


def run_testfile(path: AnyStr, file: AnyStr, app_model: SmartAppModel, settings: Settings, user_cls: type,
                 parametrizer_cls: type, storaged_predefined_fields: Dict[str, Any],
                 interactive: bool = False) -> Tuple[int, int]:
    test_file_path = os.path.join(path, file)
    if not os.path.isfile(test_file_path) or not test_file_path.endswith('.json'):
        raise FileNotFoundError
    with open(test_file_path, "r") as test_file:
        json_obj = json.load(test_file)
        success = 0
        for test_case in json_obj:
            test_params = json_obj[test_case]
            if isinstance(test_params, list):
                test_params = {"messages": test_params, "user": {}}
            print(f"[+] Processing test case {test_case} from {test_file_path}")
            if TestCase(
                    app_model,
                    settings,
                    user_cls,
                    parametrizer_cls,
                    **test_params,
                    storaged_predefined_fields=storaged_predefined_fields,
                    interactive=interactive,
            ).run():
                print(f"[+] {test_case} OK")
                success += 1
        print(f"[+] {file} {success}/{len(json_obj)}")
    return len(json_obj), success


class TestSuite:
    def __init__(self, path: AnyStr, app_config: Any, predefined_fields_storage: AnyStr):
        self.path = path
        self.app_config = app_config
        with open(predefined_fields_storage, "r") as f:
            self.storaged_predefined_fields = json.load(f)

    @lazy
    def app_model(self):
        return self.app_config.MODEL(
            self.resources, self.app_config.DIALOGUE_MANAGER, custom_settings=self.settings,
            app_name=self.app_config.APP_NAME
        )

    @lazy
    def resources(self):
        source = self.settings.get_source()

        return self.app_config.RESOURCES(source, self.app_config.REFERENCES_PATH, self.settings)

    @lazy
    def settings(self):
        return self.app_config.SETTINGS(config_path=self.app_config.CONFIGS_PATH,
                                        secret_path=self.app_config.SECRET_PATH,
                                        references_path=self.app_config.REFERENCES_PATH,
                                        app_name=self.app_config.APP_NAME)

    def run(self):
        total = 0
        total_success = 0
        for path, dirs, files in os.walk(self.path):
            for file in files:
                if not file.endswith(".json"):
                    continue
                file_total, file_success = run_testfile(
                    path,
                    file,
                    self.app_model,
                    self.settings,
                    self.app_config.USER,
                    self.app_config.PARAMETRIZER,
                    self.storaged_predefined_fields
                )
                total += file_total
                total_success += file_success

        print(f"[+] Total: {total_success}/{total}")


class TestCase:
    def __init__(self, app_model: SmartAppModel, settings: Settings, user_cls: type, parametrizer_cls: type,
                 messages: dict, storaged_predefined_fields: Dict[str, Any], interactive: bool,
                 user: Optional[dict] = None):
        self.messages = messages
        self.user_state = json.dumps(user)
        self.interactive = interactive

        self.app_model = app_model
        self.settings = settings
        self.storaged_predefined_fields = storaged_predefined_fields

        self.__parametrizer_cls = parametrizer_cls
        self.__user_cls = user_cls

    def run(self) -> bool:
        success = True

        app_callback_id = None
        for index, message in enumerate(self.messages):
            if index and self.interactive:
                print("Нажмите ENTER, чтобы продолжить...")
                input()

            request = message["request"]
            response = message["response"]

            # Если использован флаг linkPreviousByCallbackId и после предыдущего сообщения был сохранен app_callback_id,
            # сообщению добавляются заголовки. Таким образом, сработает behavior, созданный предыдущим запросом
            if message.get(LINK_BEHAVIOR_FLAG) and app_callback_id:
                headers = [(CALLBACK_ID_HEADER, app_callback_id.encode())]
            else:
                headers = [('kafka_correlationId', 'test_123')]
            message = self.create_message(request, headers=headers)

            user = self.__user_cls(
                id=message.uid, message=message, db_data=self.user_state, settings=self.settings,
                descriptions=self.app_model.scenario_descriptions,
                parametrizer_cls=self.__parametrizer_cls
            )

            commands = self.app_model.answer(message, user) or []

            answers = self._generate_answers(
                user=user, commands=commands, message=message
            )

            predefined_fields_resp = response.get("predefined_fields")
            if predefined_fields_resp:
                response = self.handle_predefined_fields_response(predefined_fields_resp, response)
            expected_answers = response["messages"]
            expected_user = response["user"]

            if len(commands) != len(response["messages"]):
                print(
                    f"[!] Expected quantity of mesages differ from received.\n"
                    f" Expected: {len(response['messages'])}. Actual: {len(answers)}."
                )
                success = False
                continue

            app_callback_id = None
            for actual, expected in zip(answers, expected_answers):
                actual_value = actual.as_dict
                diff = partial_diff(expected, actual_value)
                if diff:
                    success = False
                    print(diff)
                # Последний app_callback_id в answers, испольуется в заголовках следующего сообщения
                app_callback_id = actual.request.values.get(CALLBACK_ID_HEADER, app_callback_id)

            user_diff = partial_diff(expected_user, user.raw)
            if user_diff:
                success = False
                print(user_diff)
            self.user_state = user.raw_str
        return success

    def _generate_answers(self, user, commands, message):
        answers = []
        commands = commands or []

        commands = combine_commands(commands, user)

        for command in commands:
            request = SmartKitKafkaRequest(id=None, items=command.request_data)
            answer = SmartAppToMessage(command=command, message=message, request=request)
            answers.append(answer)
        return answers

    def create_message(self, data, headers=None):
        defaults = Environment().as_dict

        predefined_fields = data.get("predefined_fields")
        is_payload_field = data.get("payload")
        if predefined_fields:
            predefined_fields_data = self.storaged_predefined_fields[predefined_fields]
            if not is_payload_field and not predefined_fields_data.get("payload"):
                predefined_fields_data = {"payload": predefined_fields_data}
            if is_payload_field and predefined_fields_data.get("payload"):
                raise Exception("Payload field is in test case and in predefined_fields object, check it!")
            defaults.update(predefined_fields_data)
            del data["predefined_fields"]

        message = data.get("message")
        if message:
            defaults["payload"].update({"message": message})

        defaults.update(data)
        return SmartAppFromMessage(json.dumps(defaults), headers=headers)

    def handle_predefined_fields_response(self, predefined_fields_resp, response):
        predefined_fields_resp_data = self.storaged_predefined_fields[predefined_fields_resp]
        response.update(predefined_fields_resp_data)
        del response["predefined_fields"]

        pronounce_texts = response.get("pronounce_texts", [])
        if pronounce_texts:
            for text, msg_dict in zip(pronounce_texts, response["messages"]):
                msg_dict["payload"]["pronounceText"] = text
            del response["pronounce_texts"]

        return response



from core.logging.logger_utils import log
from core.text_preprocessing.preprocessing_result import TextPreprocessingResult
import scenarios.logging.logger_constants as log_const
from scenarios.actions.action import ClearCurrentScenarioAction

from smart_kit.handlers.handler_base import HandlerBase


class HandlerCloseApp(HandlerBase):
    def __init__(self, app_name):
        super(HandlerCloseApp, self).__init__(app_name)
        self._clear_current_scenario = ClearCurrentScenarioAction(None)

    def run(self, payload, user):
        super().run(payload, user)
        text_preprocessing_result = TextPreprocessingResult(payload.get("message", {}))
        params = {
            log_const.KEY_NAME: "HandlerCloseApp"
        }
        self._clear_current_scenario.run(user, text_preprocessing_result)
        log("HandlerCloseApp with text preprocessing result: '{}'".format(text_preprocessing_result.raw),
                      user, params)



import logging
from typing import List, Tuple, Optional

from confluent_kafka.cimpl import Message
from jaeger_client import SpanContext
from jaeger_client.codecs import TextCodec
from jaeger_client.constants import TRACE_ID_HEADER, BAGGAGE_HEADER_PREFIX, DEBUG_ID_HEADER_KEY, BAGGAGE_HEADER_KEY

KAFKA_MAP = "kafka_map"


class KafkaCodec(TextCodec):
    """
    Кодек предназначен для серилизации/десерелизации трейса в/из заголовок(а) кафки
    """
    VALUE_POS = 1
    KEY_POS = 0
    ENCODING = 'utf-8'

    def __init__(self,
                 url_encoding=False,
                 trace_id_header=TRACE_ID_HEADER,
                 baggage_header_prefix=BAGGAGE_HEADER_PREFIX,
                 debug_id_header=DEBUG_ID_HEADER_KEY,
                 baggage_header=BAGGAGE_HEADER_KEY):

        super().__init__(url_encoding, trace_id_header, baggage_header_prefix, debug_id_header, baggage_header)

        self.target_keys = {self.trace_id_header, self.baggage_prefix, self.debug_id_header, self.baggage_header}
        self.__logger = logging.getLogger(__name__)

    def inject(self, span_context, carrier: List[Tuple[str, bytes]]):
        if carrier is None:
            carrier = []
        existing_keys = {}
        for pos, el in enumerate(carrier):
            existing_keys[el[KafkaCodec.KEY_POS]] = pos

        carrier_dict = {}
        super().inject(span_context, carrier_dict)

        for key, value in carrier_dict.items():
            if key in existing_keys:
                carrier[existing_keys[key]] = (key, value.encode(KafkaCodec.ENCODING))
            else:
                carrier.append((key, value.encode(KafkaCodec.ENCODING)))

    def extract(self, carrier: Message) -> Optional[SpanContext]:
        header = carrier.headers()
        header_dict = {}
        if header:
            for el in header:
                if el[KafkaCodec.KEY_POS] in self.target_keys:
                    header_dict[el[KafkaCodec.KEY_POS]] = el[KafkaCodec.VALUE_POS].decode(KafkaCodec.ENCODING)
        span_context: SpanContext = None
        try:
            span_context = super().extract(header_dict)
        except Exception:
            self.__logger.exception("Could not extract SpanContext. Using new one")

        return span_context



from core.descriptions.smart_updatable_lazy_descriptions import SmartUpdatableLazyDescriptions
from scenarios.scenario_models.field.field_filler_description import field_filler_factory


class ExternalFieldFillerDescriptions(SmartUpdatableLazyDescriptions):
    def __init__(self, items):
        super(ExternalFieldFillerDescriptions, self).__init__(field_filler_factory, items, ordered=True)



from core.request.base_request import BaseRequest


class KafkaRequest(BaseRequest):
    TOPIC_KEY = "topic_key"
    KAFKA_KEY = "kafka_key"

    def __init__(self, items, id=None):
        super(KafkaRequest, self).__init__(items)
        items = items or {}
        self.topic_key = items.get(self.TOPIC_KEY)
        self.kafka_key = items.get(self.KAFKA_KEY)

    def update_empty_items(self, items):
        self.topic_key = self.topic_key or items["topic_key"]
        self.kafka_key = self.kafka_key or items["kafka_key"]

    @property
    def group_key(self):
        return "{}_{}".format(self.kafka_key, self.topic_key) if (self.topic_key and self.kafka_key) else None

    def _get_new_headers(self, source_mq_message):
        headers = source_mq_message.headers() or []
        return headers

    def run(self, data, params=None):
        publishers = params["publishers"]
        publisher = publishers[self.kafka_key]
        source_mq_message = params["mq_message"]
        publisher.send(data, source_mq_message.key(), self.topic_key, headers=self._get_new_headers(source_mq_message))

    def __str__(self):
        return f"KafkaRequest: topic_key={self.topic_key} kafka_key={self.kafka_key}"



# coding: utf-8


class Registered(dict):

    def __getitem__(self, key):
        value = self.get(key, key)
        assert not isinstance(value, str), "{} factory is not registered".format(key)
        return value


registered_factories = Registered()



from core.repositories.items_repository import ItemsRepository
from core.logging.logger_utils import log
import core.logging.logger_constants as log_const


class FileRepository(ItemsRepository):
    def __init__(self, filename, loader, source=None, save_target=None, saver=None, *args, **kwargs):
        super(FileRepository, self).__init__(source=source, *args, **kwargs)
        self.filename = filename
        self.loader = loader
        self.saver = saver
        self.save_target = save_target

    def load(self):
        if self.source.path_exists(self.filename):
            with self.source.open(self.filename, 'rb') as stream:
                binary_data = stream.read()
                data = binary_data.decode()
                self.fill(self.loader(data))
        else:
            params = {
                "error_repository_path": self.filename,
                log_const.KEY_NAME: log_const.EXCEPTION_VALUE
            }
            log("FileRepository.load loading failed with file %(error_repository_path)s",
                          params=params, level="WARNING")
        super(FileRepository, self).load()

    def save(self, save_parameters):
        with self.source.open(self.save_target, 'wb') as stream:
            stream.write(self.saver(self.data, **save_parameters).encode())



# coding: utf-8
import unittest
from unittest.mock import Mock
from scenarios.behaviors import behavior_descriptions


class BehaviorsTest2(unittest.TestCase):
    def setUp(self):
        self.test_items = {"success_action": []}

    def test_behavior_descriptions_init(self):
        obj = behavior_descriptions.BehaviorDescriptions(self.test_items)
        # self.assertTrue(hasattr(obj, 'misstate'))
        self.assertTrue(hasattr(obj, 'update_data'))



# coding: utf-8
import json
from lazy import lazy

from core.model.queued_objects.limited_queued_hashable_objects_description import \
    LimitedQueuedHashableObjectsDescription
from core.model.queued_objects.limited_queued_hashable_objects import LimitedQueuedHashableObjects, \
    LimitedQueuedHashableObjectsItems
from core.logging.logger_utils import log
from core.model.field import Field
from core.model.model import Model
from core.basic_models.parametrizers.parametrizer import BasicParametrizer
from core.basic_models.counter.counters import Counters
from core.basic_models.variables.variables import Variables


class BaseUser(Model):
    USER_DB_VERSION = "user_db_data_version"

    counters: Counters
    variables: Variables
    local_vars: Variables

    def __init__(self, id, message, values, descriptions, load_error=False):
        self.id = id
        self.message = message
        self.descriptions = descriptions
        self.load_error = load_error
        super(BaseUser, self).__init__(values=values, user=self)
        self._initialize()

    def _initialize(self):
        """Add here some things to do after creating fields."""
        pass

    @lazy
    def parametrizer(self):
        return BasicParametrizer(self, {})

    @property
    def fields(self):
        return [
            Field("counters", Counters),
            Field(
                "last_action_ids",
                LimitedQueuedHashableObjectsItems,
                self.descriptions["last_action_ids"]
            ),
            Field(
                "last_messages_ids",
                LimitedQueuedHashableObjects,
                LimitedQueuedHashableObjectsDescription(None)
            ),
            Field("local_vars", Variables, None, False),
            Field("variables", Variables),
        ]

    @property
    def raw(self):
        log("%(class_name)s.raw USER %(uid)s SAVE db_version = %(db_version)s.", self,
            {"db_version": str(self.variables.get(self.USER_DB_VERSION)),
             "uid": str(self.id)})
        return super(BaseUser, self).raw

    @property
    def raw_str(self):
        return json.dumps(self.raw)

    def expire(self):
        self.counters.expire()
        self.variables.expire()
        self.local_vars.expire()





MASK = "***"
DEFAULT_MASKING_FIELDS = ["token", "access_token", "refresh_token", "epkId", "profileId"]


def masking(data, masking_fields=None):
    if masking_fields is None:
        masking_fields = DEFAULT_MASKING_FIELDS
    if hasattr(data, 'items'):
        for key in data:
            if key in masking_fields:
                data[key] = MASK
            elif isinstance(data[key], dict):
                masking(data[key], masking_fields)
            elif isinstance(data[key], list):
                for item in data[key]:
                    masking(item, masking_fields)



from core.logging.logger_utils import log

from core.text_preprocessing.preprocessing_result import TextPreprocessingResult
import scenarios.logging.logger_constants as log_const

from smart_kit.handlers.handler_base import HandlerBase
from smart_kit.message.app_info import AppInfo
from core.names.field import APP_INFO
from smart_kit.names.message_names import MESSAGE_TO_SKILL, RUN_APP, SERVER_ACTION
from smart_kit.utils.monitoring import smart_kit_metrics


class HandlerRespond(HandlerBase):
    handler_name = "HandlerRespond"

    def __init__(self, app_name, action_name=None, ):
        super(HandlerRespond, self).__init__(app_name)
        self._action_name = action_name

    def get_action_name(self, payload, user):
        return self._action_name

    def get_action_params(self, payload):
        return {}

    def run(self, payload, user):
        callback_id = user.message.callback_id
        if user.behaviors.has_callback(callback_id):
            params = {log_const.KEY_NAME: "handling_respond"}
            log("HandlerRespond started", user, params)
            action_params = user.behaviors.get_callback_action_params(callback_id)
            if action_params:
                app_info = None
                for original_message_name in [MESSAGE_TO_SKILL, SERVER_ACTION, RUN_APP]:
                    if original_message_name in action_params:
                        app_info = AppInfo(action_params[original_message_name].get(APP_INFO, {}))
                        break

                smart_kit_metrics.counter_incoming(self.app_name, user.message.message_name, self.__class__.__name__,
                                                   user, app_info=app_info)

        text_preprocessing_result = TextPreprocessingResult(payload.get("message", {}))
        params = {
            log_const.KEY_NAME: log_const.NORMALIZED_TEXT_VALUE,
            "normalized_text": str(text_preprocessing_result.raw),
        }
        log("text preprocessing result: '%(normalized_text)s'", user, params)
        action_name = self.get_action_name(payload, user)
        action = user.descriptions["external_actions"][action_name]
        action_params = self.get_action_params(payload)
        return action.run(user, text_preprocessing_result, action_params)



# coding: utf-8
from prometheus_client import Counter, Histogram


class Monitoring:
    COUNTER = "counter"
    HISTOGRAM = "histogram"
    DEFAULT_ENABLED = True

    def __init__(self):
        self._enabled = self.DEFAULT_ENABLED
        self._monitoring_items = {self.COUNTER: {},
                                  self.HISTOGRAM: {}
                                  }

    def check_enabled(self):
        return self._enabled

    def turn_on(self):
        self._enabled = True

    def turn_off(self):
        self._enabled = False

    def got_counter(self, name):
        if self.check_enabled():
            counter = self._monitoring_items[self.COUNTER]
            if not counter.get(name):
                counter[name] = Counter(name, name)
            counter[name].inc()

    def got_histogram(self, name):
        if self.check_enabled():
            histogram = self._monitoring_items[self.HISTOGRAM]
            if not histogram.get(name):
                histogram[name] = Histogram(name, name)
            return histogram[name].time()

    def got_histogram_observe(self, name, value):
        if self.check_enabled():
            histogram = self._monitoring_items[self.HISTOGRAM]
            if not histogram.get(name):
                histogram[name] = Histogram(name, name)
            return histogram[name].observe(value)


monitoring = Monitoring()



# coding: utf-8
from core.logging.logger_utils import log
from core.names import field

import scenarios.logging.logger_constants as log_const
from lazy import lazy

from smart_kit.system_answers.nothing_found_action import NothingFoundAction
from smart_kit.utils.monitoring import smart_kit_metrics


class DialogueManager:
    NOTHING_FOUND_ACTION = "nothing_found_action"

    def __init__(self, scenario_descriptions, app_name, **kwargs):
        log(f"{self.__class__.__name__}.__init__ started.",
                      params={log_const.KEY_NAME: log_const.STARTUP_VALUE})
        self.scenario_descriptions = scenario_descriptions
        self.scenarios = scenario_descriptions['scenarios']
        self.scenario_keys = set(self.scenarios.get_keys())
        self.actions = scenario_descriptions["external_actions"]
        self.app_name = app_name
        log("DialogueManager.__init__ finished.", params={log_const.KEY_NAME: log_const.STARTUP_VALUE})

    @lazy
    def _nothing_found_action(self):
        return self.actions.get(self.NOTHING_FOUND_ACTION) or NothingFoundAction()

    def run(self, text_preprocessing_result, user):
        last_scenarios = user.last_scenarios
        scenarios_names = last_scenarios.scenarios_names
        scenario_key = user.message.payload[field.INTENT]

        if scenario_key in scenarios_names:
            scenario = self.scenarios[scenario_key]
            if not scenario.text_fits(text_preprocessing_result, user):
                smart_kit_metrics.counter_nothing_found(self.app_name, scenario_key, user)

                return self._nothing_found_action.run(user, text_preprocessing_result), False

        return self.run_scenario(scenario_key, text_preprocessing_result, user), True

    def run_scenario(self, scen_id, text_preprocessing_result, user):
        initial_last_scenario = user.last_scenarios.last_scenario_name
        scenario = self.scenarios[scen_id]
        params = {log_const.KEY_NAME: log_const.CHOSEN_SCENARIO_VALUE,
                  log_const.CHOSEN_SCENARIO_VALUE: scen_id,
                  log_const.SCENARIO_DESCRIPTION_VALUE: scenario.scenario_description
                  }
        log(log_const.LAST_SCENARIO_MESSAGE, user, params)
        run_scenario_result = scenario.run(text_preprocessing_result, user)

        actual_last_scenario = user.last_scenarios.last_scenario_name
        if actual_last_scenario and actual_last_scenario != initial_last_scenario:
            smart_kit_metrics.counter_scenario_change(self.app_name, actual_last_scenario, user)

        return run_scenario_result



import unittest
from unittest.mock import Mock

from core.jaeger_custom_client.kafka_codec import KafkaCodec

NLP_BAGGAGE = "nlp-baggage-"

NLP_TRACE_ID = "nlp-trace-id"


class KafkaCodecTest(unittest.TestCase):

    def setUp(self) -> None:
        self.codec = KafkaCodec(trace_id_header=NLP_TRACE_ID, baggage_header_prefix=NLP_BAGGAGE)

    def test_extract(self):
        # Won't fail if there is a filter in KafkaCode extract method
        trace_id = 15441712733567627997
        span_id = 957657810971653503
        parent_id = 0
        flags = 1

        header = [('d', b'D'), ('cp', b'\xff'), ("b'0xc9'", b'0xc9'), ('q', b''), ('a', b'g '),
                  (NLP_TRACE_ID, b'd64bfbeadfecb2dd:d4a48b48757f57f:0:1')]
        carrier = Mock()
        carrier.headers.return_value = header

        actual = self.codec.extract(carrier)
        self.assertEqual(actual.trace_id, trace_id)
        self.assertEqual(actual.span_id, span_id)
        self.assertIsNone(actual.parent_id)
        self.assertEqual(actual.flags, flags)

    def test_extract_with_malformed_trace_id(self):
        header = [('d', b'D'), ('cp', b'\xff'), ("b'0xc9'", b'0xc9'), ('q', b''), ('a', b'g '),
                  (NLP_TRACE_ID, b'this_is_malformed_trace_id')]
        carrier = Mock()
        carrier.headers.return_value = header

        actual = self.codec.extract(carrier)
        self.assertIsNone(actual)

    def test_inject(self):
        span_context = Mock()
        span_context.trace_id = 15441712733567627997
        span_context.span_id = 957657810971658888
        span_context.parent_id = 957657810971653503
        span_context.flags = 1
        span_context.baggage = {"some_key": "56"}
        baggage_key = f"{NLP_BAGGAGE}test"
        header = [('d', b'D'), ('cp', b'\xff'), ("b'0xc9'", b'0xc9'), ('q', b''),
                  (NLP_TRACE_ID, b'd64bfbeadfecb2dd:d4a48b48757f57f:0:1'), ('a', b'g '),
                  (baggage_key, b'test_data')]

        expected_header = [('d', b'D'), ('cp', b'\xff'), ("b'0xc9'", b'0xc9'), ('q', b''),
                           (NLP_TRACE_ID, b'd64bfbeadfecb2dd:d4a48b487580a88:d4a48b48757f57f:1'), ('a', b'g '),
                           (baggage_key, b'test_data'),
                           (f"{NLP_BAGGAGE}some_key", b"56")]
        self.codec.inject(span_context, header)
        self.assertSequenceEqual(expected_header, header)





from setuptools import find_packages, setup
import versioneer


with open("README.md", "r", encoding="utf-8") as file:
    long_description = file.read()

setup(
    name="smart_app_framework",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author="SberDevices",
    author_email="developer@sberdevices.ru",
    description="SmartApp Framework — это фреймворк, "
                "который позволяет создавать смартапы "
                "с поддержкой виртуальных ассистентов Салют.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="sberpl-2",
    packages=find_packages(exclude=[]),
    include_package_data=True,
    install_requires=[
        "aiohttp==3.7.3",
        "aioredis",
        "boto==2.49.0",
        "confluent_kafka==1.1.0",
        "croniter",
        "dawg==0.8.0",
        "dill==0.3.3",
        "h5py<3.0.0",
        "ics==0.6",
        "jaeger_client==4.3.0",
        "Jinja2==2.10.1",
        "keras==2.3.1",
        "lazy",
        "nltk==3.5",
        "numpy",
        "objgraph==3.4.1",
        "prometheus-client==0.7.1",
        "psutil==5.8.0",
        "pyignite>=0.3.4",
        "pymorphy2==0.8",
        "pymorphy2_dicts==2.4.393442.3710985",
        "python-dateutil==2.7.3",
        "python-json-logger==0.1.11",
        "PyYAML==5.3",
        "redis",
        "requests==2.22.0",
        "rnnmorph==0.4.0",
        "rusenttokenize==0.0.5",
        "scikit-learn==0.24.1",
        "setuptools",
        "tabulate",
        "tatsu==4.4.0",
        "tensorflow==1.15",
        "timeout-decorator==0.4.1",
        "tqdm",
        "Twisted"
    ],
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7"
    ]
)



# coding: utf-8
from smart_kit.utils.monitoring import smart_kit_metrics


class HandlerBase:
    TOPIC_KEY = "template_app"
    KAFKA_KEY = "main"

    def __init__(self, app_name):
        self.app_name = app_name

    def run(self, payload, user):
        # отправка события о входящем сообщении в систему мониторинга
        smart_kit_metrics.counter_incoming(self.app_name, user.message.message_name, self.__class__.__name__,
                                           user, app_info=user.message.app_info)



# -*- coding: utf-8 -*-
import unittest
from unittest.mock import Mock
from prometheus_client import Counter, Histogram

from core.monitoring.monitoring import Monitoring


class MonitoringTest1(unittest.TestCase):
    def setUp(self):
        self.logger = Mock()
        self.logger.exception = Mock()
        self.config = Mock()
        self.mock_rep = Mock()
        self.monitoring = Monitoring()

    def test_got_message_disabled(self):
        self.monitoring.turn_off()
        self.monitoring._monitoring_items[ self.monitoring.COUNTER ] = {}
        event_name = "test_counter"
        counter_item = self.monitoring._monitoring_items[ self.monitoring.COUNTER ]
        self.assertTrue(counter_item == dict())
        counter = self.monitoring.got_histogram(event_name)
        self.assertTrue(event_name not in counter_item)
        self.assertIsNone(counter)

    def test_got_message_enabled(self):
        self.monitoring.turn_on()
        self.monitoring._monitoring_items[ self.monitoring.COUNTER ] = {}
        event_name = "test_counter"
        counter_item = self.monitoring._monitoring_items[ self.monitoring.COUNTER ]
        self.assertTrue(counter_item == dict())
        self.monitoring.got_counter(event_name)
        self.assertTrue(event_name in counter_item)
        self.assertEqual(type(counter_item[ event_name ]), type(Counter('counter_name', 'counter_name')))

    def test_got_histogram_disabled(self):
        self.monitoring.turn_off()
        self.monitoring._monitoring_items[ self.monitoring.HISTOGRAM ] = {}
        event_name = "test_histogram"
        histogram_item = self.monitoring._monitoring_items[ self.monitoring.HISTOGRAM ]
        self.assertTrue(histogram_item == dict())
        histogram = self.monitoring.got_histogram(event_name)
        self.assertTrue(event_name not in histogram_item)
        self.assertIsNone(histogram)

    def test_got_histogram_enabled(self):
        self.monitoring.turn_on()
        self.monitoring._monitoring_items[self.monitoring.HISTOGRAM] = {}
        event_name = "test_histogram"
        histogram_item = self.monitoring._monitoring_items[self.monitoring.HISTOGRAM]
        self.assertTrue(histogram_item == dict())
        self.monitoring.got_histogram(event_name)
        histogram_item = self.monitoring._monitoring_items[self.monitoring.HISTOGRAM]
        self.assertTrue(event_name in histogram_item)
        self.assertEqual(type(histogram_item[event_name]), type(Histogram('histogram_name', 'histogram_name')))



# coding: utf-8
from lazy import lazy

from core.model.factory import factory, list_factory
from core.basic_models.actions.basic_actions import Action
from scenarios.scenario_models.field_requirements.field_requirements import FieldRequirement


class FieldValidator:
    def __init__(self, items):
        self._requirement = items.get("requirement")
        self._actions = items.get("actions")

    @lazy
    @factory(FieldRequirement)
    def requirement(self):
        return self._requirement

    @lazy
    @list_factory(Action)
    def actions(self):
        return self._actions



from collections import Iterable, Mapping
import re


class LogMasker:
    """
    Regular expression is used to find 16 digit or 18 digit number that could be
    between single or double quotes
    and any non digit character
    """
    card_regular = re.compile("(?:^|\s|\'|\")(?:(\d{18})|(\d{16})|(?:\d{4} ){3}(\d{4})(\s?\d{2})?)(?:$|\D)")
    d_regular = re.compile("\d")

    @classmethod
    def _card_sub_func(cls, x):
        g0 = x.group(0)

        is_last_not_digit = int(g0 and not g0[-1].isdigit())
        last_char = g0[-1]

        mask = cls.d_regular.sub("*", x.group(0))[:-(4 + is_last_not_digit)]
        digs = ((x.group(1) or '') + (x.group(2) or '') + (x.group(3) or '') + (x.group(4) or '')).replace(' ', '')[-4:]
        return mask + digs + (last_char * is_last_not_digit)

    @classmethod
    def mask_structure(cls, record, func):
        if isinstance(record, Mapping):
            return {k: cls.mask_structure(v, func) for k, v in record.items()}
        elif isinstance(record, str):
            return func(record)
        elif isinstance(record, Iterable):
            return [cls.mask_structure(i, func) for i in record]
        else:
            return record

    @classmethod
    def regular_exp(cls, record):
        return cls.card_regular.sub(cls._card_sub_func, record)

    @classmethod
    def percent_fix(cls, record):
        return str(record).replace("%", "%%")



from core.logging.logger_constants import KEY_NAME
from core.logging.logger_utils import log
from core.monitoring.monitoring import monitoring

from prometheus_client import Counter


def _filter_monitoring_msg(msg):
    msg = msg.replace("-", ":")
    return msg


def silence_it(func):
    def wrap(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except:
            log("Metrics: Failed send. Exception occurred.",
                          params={KEY_NAME: "metrics_fail"}, level="ERROR", exc_info=True)
    return wrap


class Metrics:

    @silence_it
    def init_metrics(self, app_name):
        self._get_or_create_counter(_filter_monitoring_msg("{}_load_error".format(app_name)), "Load user data error")
        self._get_or_create_counter(_filter_monitoring_msg("{}_save_error".format(app_name)), "Save user data error")
        self._get_or_create_counter(_filter_monitoring_msg("{}_save_collision".format(app_name)), "Save user data collision")
        self._get_or_create_counter(_filter_monitoring_msg("{}_save_collision_tries_left".format(app_name)),
                                            "Save user data collision all retries left.")
        self._get_or_create_counter(_filter_monitoring_msg("{}_exception".format(app_name)), "Exception in run-time.")
        self._get_or_create_counter(_filter_monitoring_msg("{}_invalid_message".format(app_name)),
                                            "Incoming message validation error.")

    def _get_or_create_counter(self, monitoring_msg, descr, labels=()):
        counter = getattr(self, monitoring_msg, None)
        if not counter:
            counter = Counter(monitoring_msg, descr, labels)
            setattr(self, monitoring_msg, counter)
        return counter

    @silence_it
    def counter_incoming(self, app_name, message_name, handler, user, app_info=None):
        monitoring_msg = _filter_monitoring_msg("{}_incoming".format(app_name))

        c = self._get_or_create_counter(monitoring_msg, "Count of incoming messages",
                                        ['message_name', 'handler', 'project_id', 'system_name', 'application_id',
                                         'app_version_id', 'channel', 'surface'])

        if app_info is not None:
            project_id = app_info.project_id
            system_name = app_info.system_name
            application_id = app_info.application_id
            app_version_id = app_info.app_version_id
        else:
            project_id = system_name = application_id = app_version_id = None
        c.labels(message_name, handler, project_id, system_name,
                 application_id, app_version_id, user.message.channel, user.message.device.surface).inc()

    @silence_it
    def counter_outgoing(self, app_name, message_name, outgoing_message, user):

        monitoring_msg = _filter_monitoring_msg("{}_outgoing".format(app_name))

        c = self._get_or_create_counter(monitoring_msg, "Count of outgoing requests from application.",
                                        ['message_name', 'project_id', 'system_name', 'application_id',
                                         'app_version_id', 'channel', 'surface'])
        app_info = user.message.app_info
        c.labels(message_name, app_info.project_id, app_info.system_name,
                 app_info.application_id,
                 app_info.app_version_id, user.message.channel, user.message.device.surface).inc()

    @silence_it
    def counter_scenario_change(self, app_name, scenario, user):
        monitoring_msg = "{}_scenario_change".format(app_name)

        c = self._get_or_create_counter(monitoring_msg, "Count of scenario change events",
                                        ['scenario', 'project_id', 'system_name', 'application_id',
                                         'app_version_id', 'channel', 'surface'])
        app_info = user.message.app_info
        c.labels(scenario, app_info.project_id, app_info.system_name,
                 app_info.application_id,
                 app_info.app_version_id, user.message.channel, user.message.device.surface).inc()

    @silence_it
    def counter_nothing_found(self, app_name, scenario, user):
        monitoring_msg = "{}_outgoing_nothing_found".format(app_name)
        c = self._get_or_create_counter(monitoring_msg, "Count of scenario nothing found events",
                                        ['scenario', 'project_id', 'system_name', 'application_id',
                                         'app_version_id', 'channel', 'surface'])
        app_info = user.message.app_info
        c.labels(scenario, app_info.project_id, app_info.system_name,
                 app_info.application_id,
                 app_info.app_version_id, user.message.channel, user.message.device.surface).inc()

    @silence_it
    def counter_load_error(self, app_name):
        monitoring_msg = "{}_load_error".format(app_name)
        c = self._get_or_create_counter(_filter_monitoring_msg(monitoring_msg), "Load user data error")
        c.inc()

    @silence_it
    def counter_save_error(self, app_name):
        monitoring_msg = "{}_save_error".format(app_name)
        c = self._get_or_create_counter(_filter_monitoring_msg(monitoring_msg), "Save user data error")
        c.inc()

    @silence_it
    def counter_save_collision(self, app_name):
        monitoring_msg = "{}_save_collision".format(app_name)
        c = self._get_or_create_counter(_filter_monitoring_msg(monitoring_msg), "Save user data collision")
        c.inc()

    @silence_it
    def counter_save_collision_tries_left(self, app_name):
        monitoring_msg = "{}_save_collision_tries_left".format(app_name)
        c = self._get_or_create_counter(_filter_monitoring_msg(monitoring_msg), "Save user data collision all retries left.")
        c.inc()

    @silence_it
    def counter_exception(self, app_name):
        monitoring_msg = "{}_exception".format(app_name)
        c = self._get_or_create_counter(_filter_monitoring_msg(monitoring_msg), "Exception in run-time.")
        c.inc()

    @silence_it
    def counter_invalid_message(self, app_name):
        monitoring_msg = "{}_invalid_message".format(app_name)
        c = self._get_or_create_counter(_filter_monitoring_msg(monitoring_msg), "Incoming message validation error.")
        c.inc()

    @silence_it
    def counter_behavior_success(self, app_name, request_message_name):
        # 'Number of Success replies on messageName'
        self._behavior_monitoing_by_status(app_name, "success", request_message_name)

    @silence_it
    def counter_behavior_fail(self, app_name, request_message_name):
        # 'Number of Fail replies on messageName'
        self._behavior_monitoing_by_status(app_name, "fail", request_message_name)

    @silence_it
    def counter_behavior_misstate(self, app_name, request_message_name):
        # 'Number of Misstate replies on messageName'
        self._behavior_monitoing_by_status(app_name, "misstate", request_message_name)

    @silence_it
    def counter_behavior_timeout(self, app_name, request_message_name):
        # 'Number of Timeout replies on messageName'
        self._behavior_monitoing_by_status(app_name, "timeout", request_message_name)

    @silence_it
    def counter_behavior_expire(self, app_name, request_message_name):
        # 'Number of expire events on messageName'
        self._behavior_monitoing_by_status(app_name, "expire", request_message_name)

    def _behavior_monitoing_by_status(self, app_name, status, request_message_name):
        monitoring_msg = '{}_callback'.format(app_name)
        c = self._get_or_create_counter(monitoring_msg,
                                        "Count of incoming callback events with request_message_name",
                                        ['request_message_name', 'status'])

        c.labels(request_message_name, status).inc()

    @silence_it
    def counter_mq_long_waiting(self, app_name):
        monitoring_msg = "{}_mq_long_waiting".format(app_name)
        c = self._get_or_create_counter(_filter_monitoring_msg(monitoring_msg), "(Now - creation_time) is greater than threshold")
        c.inc()

    @silence_it
    def sampling_load_time(self, app_name, value):
        monitoring_msg = "{}_load_time".format(app_name)
        monitoring.got_histogram_observe(_filter_monitoring_msg(monitoring_msg), value)

    @silence_it
    def sampling_script_time(self, app_name, value):
        monitoring_msg = "{}_script_time".format(app_name)
        monitoring.got_histogram_observe(_filter_monitoring_msg(monitoring_msg), value)

    @silence_it
    def sampling_save_time(self, app_name, value):
        monitoring_msg = "{}_save_time".format(app_name)
        monitoring.got_histogram_observe(_filter_monitoring_msg(monitoring_msg), value)

    @silence_it
    def sampling_mq_waiting_time(self, app_name, value):
        monitoring_msg = "{}_mq_waiting_time".format(app_name)
        monitoring.got_histogram_observe(_filter_monitoring_msg(monitoring_msg), value)


smart_kit_metrics = Metrics()
